#!/usr/bin/python

# Bryan Jacoby, 11/2006
# Program to calculate S/N for phased array pulsar observations at frequencies
# below 100 MHz.  Requires: programs t408 (returns Haslam et al. 408 MHz sky
# temperature for a given l,b) and NE2001 (Cordes & Lazio's galactic electron
# model) and pulsar catalog file psrcat_star.dat generated from ATNF pulsar
# catalog.


import os
import math
import re



# Uncomment one of the two mode lines below.  Table gives a full table of all pulsars,
# fluxratio gives a limited table of pulsars with measured low-frequency flux densities
# to allow assessment of low-frequency flux density extrapolation used for the pulsars
# without low-frequency observations.

mode = 'table'
#mode = 'fluxratio'


# Uncomment one of the three instruments below (only matters in 'table' mode).

inst = 'LWA1'
#inst = 'LWDA'
#inst = 'ETA'

# Change freq, tsamp, tint, fresmax, and fresmin for a given instrument to experiment with different
# observation parameters.

if inst == 'LWA1':
    # LWA1 can do 20-80 MHz
    # VLA latitude = 34d04'42.497s N
    obslat = 34.078
    freq = 74 #MHz
    tsamp = 1. #ms
    tint = 1000. #s
    fresmax = 0.0001  #Smallest possible channel BW in MHz    
    fresmin = 0. #Largest possible channel BW in MHz; 0 -> full bandwidth (bw)
    bw = 8.  #MHz
    numstands = 256
    freqsat = 37.5 #Frequency at which effective area saturates
    npol = 2

if inst == 'LWDA':
    # LWDA is 74 MHz
    # VLA latitude = 34d04'42.497s N
    obslat = 34.078
    tsamp = 1. #ms
    tint = 1000. #s
    fresmax = 0.019  #Smallest possible channel BW in MHz
    fresmin = 0.  #Largest possible channel BW in MHz; 0 -> full bandwidth (bw)
    bw = 1.6  #MHz
    numstands = 16    
    freqsat = 37.5 #Frequency at which effective area saturates
    npol = 1

if inst == 'ETA':
    # ETA can do 29-47 MHz
    # PARI latitude = 35d11.8' N
    obslat = 35.197
    freq = 44.5 #MHz
    tsamp = 10. #ms    
    tint = 10000. #s
    fresmax = 0.001  #Smallest possible channel BW in MHz
    fresmin = 0   #Largest possible channel BW in MHz; 0 -> full bandwidth (bw)
    bw = 5. #MHz 
    numstands = 12
    freqsat = 0 #Frequency at which effective area saturates, don't know value for ETA
    npol = 2

gain = 1e-26 * 0.13 * (numstands * math.pow((300./(max(freq,freqsat))),2.)) / 2. / 1.38e-23 
# gain = 1e-26 * 0.7 *  7854.  / 2. / 1.38e-23 # Check gain calculation with GBT, should be 2

if fresmin == 0:
    fresmin = bw

if mode == 'table':
# Print out header-type information
    print '# Instrument =', inst
    print '# Frequency =', freq, 'MHz'
    print '# Gain at zenith =', gain, 'K/Jy'
    print '# Sampling interval =', tsamp, 'ms'
    print '# Integration time =', tint,'s'
    print '# Maximum frequency resolution =', fresmax*1000., 'kHz'
    print '# Bandwidth =', bw, 'MHz'
    print '#----------------------------------------------------------------------------------------------------------------------------------------------------------------------------'
    print '# NAME        RAJ         DECJ                l          b         P        DM       W50      S400  SPINDX   S4m  Freq   S_pred  Tsky   Ch BW  tau_DM  tau_scat   DC     S/N'
    print '#             (hms)       (dms)            (deg)      (deg)      (ms)  (cm^-3 pc)    (ms)     (mJy)         (mJy) (MHz)   (mJy)   (K)   (kHz)  (ms)       (ms)'
    print '#----------------------------------------------------------------------------------------------------------------------------------------------------------------------------'

if mode == 'fluxratio':
    print '#Name            P       S400  SPINDX    S4m   Freq  Calc.S4m  Ratio'
    print '#               (ms)     (mJy)          (mJy)  (MHz)   (mJy)'
    print '#-------------------------------------------------------------------'

#########################################################

def gettsky(inst,l,b,freq):
# Use t408 program to get sky temperature at pulsar position.  Note that we don't pay
# attention to sky temperature elsewhere in the beam.  t408 doesn't take command
# line arguments so we make a file t408.in with the desired galactic coordinates.
    """408 MHz sky temp scaled to freq with spectral index -2.6."""
    slb=str(l) + ',' + str(b) + '\n'
    f=open('t408.in', 'w')
    f.write(slb)
    f.close()
    os.system('t408 < t408.in >& t408.out')
    f=open('t408.out', 'r')
    t408str = f.readline()
    while t408str[:6]!=' t408:':
        t408str = f.readline()

    f.close()
    t408 = float(t408str[8:])
    skytemp = t408 * math.pow((freq / 408.),-2.6)
    return skytemp

#########################################################

def getne2001tau(l,b,freq,dm):
# Use NE2001 to get scattering time.  
    """Scattering time from NE2001 with Kolgomorov frequency scaling."""
    slbdm = str(l) + ' ' + str(b) + ' ' + str(dm) + ' '
    os.chdir('/home/baj/run_NE2001')
    os.system('./run_NE2001.pl %s 1 tau > NE2001.out' % slbdm)
    f=open('NE2001.out', 'r')
    ne2001str = f.readline()

    while ne2001str[:5]!='TAU =':
        ne2001str = f.readline()

    f.close()
    tauscat1ghz = float(ne2001str[6:17])
    tauscat = tauscat1ghz * (1000. / freq)**4.4
    # Frequency scaling from Bhat et al. 2004 is 3.9
    return tauscat # pulse broadening in ms

#########################################################

def getpsrs(inst,gain,tsamp,tint,fresmax,bw):
    """Read pulsar catalog and calculate S/N for each entry."""
    count = 0
    f=open('/home/baj/lwa/psrcat_star.dat', 'r')
    catstrraw = f.readline()

    while catstrraw[:1]=="#":
        catstrraw=f.readline()

    while 1:
        catstr = re.sub('\*','0',catstrraw)
        l = float(catstr[47:54])
        b = float(catstr[58:65])
        p = float(catstr[68:76])
        if p == 0:
            catstrraw = f.readline()
            if not catstrraw:
                break
            continue
        dm = float(catstr[78:88])
        w50 = float(catstr[90:98])
        if w50 == 0:
            w50 = 0.1 * p * 1000.
        s400 = float(catstr[100:108])
        spindx = float(catstr[110:116])
# Default spindx if none in catalog
        if spindx == 0:
            spindx = -1.6
        slw = float(catstr[117:123])
        slwfreq = float(catstr[124:127])
        
        decjd = float(catstr[32:35]) + float(catstr[32:33] + catstr[36:38])/60.
# ignore arcseconds of declination to keep from messing up psrs without seconds in catalog


# Skip pulsars that never rise at telescope location
        sinmaxel = math.cos(math.fabs(obslat - decjd)*math.pi/180.)
        if sinmaxel <= 0:
            catstrraw = f.readline()
            if not catstrraw:
                break
            continue

        if ((s400 == 0)and(slw == 0)):
            catstrraw = f.readline()
            if not catstrraw:
                break
            continue

# First attempt at calculating flux density:
# Two alpha values for 61-102 MHz and 100-400 MHz calcluated according
# to Izvekova et al. 1981. except don't extrapolate beyond P range of
# their sample.
#        alpha61to102 = 2.3 * math.log10(p) - 0.3
#        if alpha61to102 < -2:
#            alpha61to102 = -2
#        alpha100to400 = 0.9 + 0.7 * math.log10(p)
#        if alpha100to400 < 0.3:
#            alpha100to400 = 0.3
#        calcs100 = s400 * math.pow((400./100.),alpha100to400)
#        calcslw = calcs100 * math.pow((100./freq),alpha61to102)
# But this Izvekova et al. method doesn't seem to work that well.

# Here is a simpler method of flux density extrapoltaion: Extrapolate from
# published flux density to 100 MHz using the published spectral index of
# assume 1.6 if no published value.  Then extrapolate from 100 MHz
# to frequency of interest with spectral index of 1.4.  This seems to
# work better than more complicated method above.
        calcs100 = s400 * math.pow((100./400.),spindx)
        calcslw = calcs100 * math.pow((100./freq),-1.4)
# If there is a low frequency flux measurement extrapolate from that instead:
        if slw != 0:
            calcslw = slw * math.pow((slwfreq/freq),-1.4)

# Throw out pulsars with no DM in catalog
        if dm > 0:
            scattime = getne2001tau(l,b,freq,dm)
        else:
            catstrraw = f.readline()
            if not catstrraw:
                break
            continue

# Calculate optimum frequency resolution, balancing DM smearing within a channel with
# the inverse bandwith/maximum fluctuation frequency of the channel.
        fresopt = math.pow(math.pow(freq,6.) / (8.3*8.3*1e18*dm*dm),0.25)
        fres = max(fresopt,fresmax)
        fres = min(fres,fresmin)
        invfres = 1.e3 * 1.e-6 / fres

# Calculate dispersion smearing time 
        dmtime = 1000. * 8300. * dm * fres / math.pow(freq,3)
# Effective pulse duty cycle
        delta =  math.sqrt(dmtime*dmtime + scattime*scattime + tsamp*tsamp + invfres*invfres + 2.*w50*2.*w50) / (p * 1000.)

        tsky = gettsky(inst,l,b,freq)

# Calculate sensitivity using modified form of radiometer equation
        if (delta < 1):
            sigma = tsky / (sinmaxel * gain) / (math.sqrt(npol * tint * bw * 1.e6)) * math.sqrt(delta / (1 - delta))
            snr = 0.001 * calcslw / sigma
        else:
            snr = 0
            
# print results
        if mode == 'table':
            print '%s %8.2f %s %8.1f %6.0f %6.3f %6.2f %10.2f %5.2f %7.2f' % (catstrraw[6:67], p * 1000., catstrraw[80:127], calcslw, tsky, fres*1000., dmtime, scattime, min(delta,1), snr)

        if ((mode == 'fluxratio')and(slw != 0)):
            calcs100comp = s400 * math.pow((100./400.),spindx)
            calcslwfreq = calcs100comp * math.pow((100./slwfreq),-1.4)
            print '%s %7.2f %8.1f %6.2f %8.1f %5.0f %7.1f %8.3f' % (catstrraw[6:18], p * 1000., s400, spindx, slw, slwfreq, calcslwfreq, calcslwfreq/slw)

        catstrraw = f.readline()
        count += 1
        if not catstrraw:
            break
    f.close()    
    return count

#########################################################


getpsrs(inst,gain,tsamp,tint,fresmax,bw)


# print notes at end of table
if mode == 'table':
    print '#----------------------------------------------------------------------------------------------------------------------------------------------------------------------------'
    print '# Columns taken from ATNF pulsar catalog: '
    print '#'
    print '# W50: FWHM of the pulse profile.  To account for the fact that profiles '
    print '#	tend to be wider at low frequencies, W50 was multiplied by 2 '
    print '#	in the program\'s calculations.'
    print '#'
    print '# S400: 400 MHz flux density.'
    print '#'
    print '# SPINDX: Published spectral index (above low-frequency spectral turnover).'
    print '#'
    print '# Columns from actual low-frequency observations:'
    print '# '
    print '# S4m: Low-frequency flux density measurements from VLSS or '
    print '#	Izvekova et al. 1981.'
    print '#'
    print '# Freq: Frequency at which S4m was measured (74 MHz = VLSS, otherwise'
    print '#	Izvekova et al. 1981).'
    print '#'
    print '# Columns calculated: '
    print '#'
    print '# S_pred: Predicted flux density at observation frequency calculated by '
    print '#	extrapolating S400 to 100 MHz using SPINDX (or assuming '
    print '#	SPINDX = -1.6 if it is not in the catalog), then extrapolating '
    print '#	from 100 MHz to observing frequency assuming a spectral index '
    print '#	of 1.4.  For pulsars with S4m, flux density was scaled from S4m '
    print '#	with a spectral index of 1.4.'
    print '#'
    print '# Tsky: Haslam et al. 408 MHz sky temperature at pulsar\'s l,b'
    print '#	extrapolated to observing frequency with a spectral index of -2.6.'
    print '#'
    print '# Ch BW: Channel bandwidth (>= maximum frequency resolution in header) giving'
    print '#	optimum effective time resolution for a given pulsar\'s DM.'
    print '#'
    print '# tau_DM: Dispersion smearing within a frequency channel.'
    print '#'
    print '# tau_scat: Pulse broadening due to scattering from NE2001 model scaled'
    print '#	by frequency^-4.4.'
    print '#'
    print '# DC: Effective (observed) duty cycle.  This accounts for W50, sampling'
    print '#	interval, inverse frequency resolution, tau_DM, and tau_scat.'
    print '#'
    print '# S/N: Expected signal-to-noise of an observation at transit with the '
    print '#	parameters in the header.'

if mode == 'fluxratio':
    print '#-------------------------------------------------------------------'
    print '# S400: Published 400 MHz flux density.'
    print '# SPINDX: Spectral index, -1.6 assumed if no published value.'
    print '# S4m: Observed flux density at ~4-meter wavelength.'
    print '# Freq: Frequency at which S4m was measured.'
    print '# Calc.S4m: Flux density at Freq calculated as for other pulsars in "table" mode.'
    print '# Ratio: Calculated/observed flux density at frequency Freq.'
